import { select } from "d3-selection";
import "d3-transition";
import { getTypeFromChartLayout } from "./utils";

function initUpdate(instance, state) {
	instance.update = function() {
		const chart_layouts = instance.chartLayouts();
		const facet_names = instance.facetNames();
		const facet_offsets = instance.facetOffsets();

		const data = chart_layouts.map(function(chart_layout, i) {
			return { chart_layout, facet_name: facet_names[i], facet_offsets: (facet_offsets[i] || {x: 0, y: 0}) };
		});

		var parseNum, dur;

		function getXTextSize() {
			return instance.layout().remToPx(state.anno_x_font_size);
		}

		function getYTextSize() {
			return instance.layout().remToPx(state.anno_y_font_size);
		}

		function splitLabelAndValue(string) {
			return string.replace(/\s*::\s*/g, "::").split("::").map(s => s.trim());
		}

		function splitRange(string) {
			return string.replace(/\s*>>\s*/, ">>").split(">>").map(s => s.trim());
		}

		function getColor(d) {
			return d.color;
		}

		function getParsedValue(label, type, chart_layout, axis) {
			var value;
			if (type === "number") value = isNaN(parseNum(label)) ? parseNum(label) : label;
			else if (type === "datetime") {
				const parser = axis === "x" ? chart_layout.xDatetimeParse() : chart_layout.yDatetimeParse();
				value = parser(label);
			}
			else value = label;
			return value;
		}

		function updateXLines(facet_group, d, stack) {
			const chart_layout = d.chart_layout;
			let margin_left = chart_layout.margins.left;
			let margin_top = chart_layout.margins.top;
			let plot_height = Math.max(0, chart_layout.plot_height);

			let a = state.anno_x_label_align;
			let v = state.anno_x_label_vertical;
			let disabled = instance.xDisabled();

			let xScale = chart_layout.xScale();
			let type = getTypeFromChartLayout(chart_layout.xData());

			var line_data = [];
			if (!disabled && state.anno_x_enabled && state.anno_x_lines && stack == state.anno_x_stack) {
				state.anno_x_lines.split("\n").forEach(function(string) {
					var line = splitLabelAndValue(string);
					var label = line[0];
					var position = line[1];
					if (position === undefined) return;
					var color = line[2] || state.anno_x_line_color;
					var facet = line[3] || null;
					if (facet && d.facet_name !== facet) return;
					line_data.push({
						label: label,
						x: getParsedValue(position, type, chart_layout, "x"),
						color: color
					});
				});
			}

			var lines = select(facet_group).selectAll("g.line.x").data(line_data);
			var lines_enter = lines.enter().append("g").attr("class", "line x")
				.attr("transform", function(d) { return "translate(" + (xScale(d.x) || margin_left) + ", " + margin_top + ")"; });
			lines_enter.append("line");
			lines_enter.append("text");
			var lines_update = lines.merge(lines_enter);
			lines_update.transition().duration(dur)
				.attr("transform", function(d) { return "translate(" + (xScale(d.x) || margin_left) + ", " + margin_top + ")"; });
			lines_update.select("line")
				.attr("stroke-dasharray", state.anno_x_line_dash + " " + state.anno_x_line_dash)
				.attr("stroke", getColor)
				.attr("stroke-width", state.anno_x_line_width)
				.attr("y2", plot_height);
			lines_update.select("text").text(function(d) { return d.label; })
				.attr("fill", state.anno_x_line_color)
				.attr("font-size", getXTextSize())
				.attr("transform", function(d) {
					var y_shift = v ? 6 : (getXTextSize() * 1.25);
					var ver = a == "bottom" ? (plot_height - y_shift) : (a == "top" ? y_shift : plot_height/2);
					if (v) return "translate(" + getXTextSize() * 0.5 + ", " + ver + ") rotate(" + (v ? -90 : 0) + ")";
					else {
						// In horizontal mode, position the annotations to the left of the line unless there's not enough
						// space. Left is chosen as the default on the basis that it's better to minimise overlapping annotations
						// at the end of the chart than the beginning, since there may be more annotations in the recent history
						// in a time series
						let rect = this.getBoundingClientRect();
						let x_shift = -getXTextSize()/2;
						if ((xScale(d.x) || margin_left) - rect.width - getXTextSize()/2 < xScale.range()[0]) x_shift += rect.width + getXTextSize();
						return "translate(" + x_shift + ", " + ver + ")";
					}
				})
				.attr("text-anchor", function() {
					if (v) return a == "bottom" ? "start" : (a == "top" ? "end" : "middle");
					return "end";
				})
				.attr("dy", function(d) {
					if (v) {
						if ((xScale(d.x) || 0) < (getXTextSize() + margin_left)) return "0.75em";
						else return "-1em";
					}
					return null;
				});
			lines.exit().remove();
		}

		function updateXAreas(facet_group, d, stack) {
			const chart_layout = d.chart_layout;

			let margin_left = chart_layout.margins.left;
			let margin_top = chart_layout.margins.top;
			let plot_height = Math.max(0, chart_layout.plot_height);
			let a = state.anno_x_label_align;
			let v = state.anno_x_label_vertical;
			let xScale = chart_layout.xScale();
			let type = getTypeFromChartLayout(d.chart_layout.xData());
			let disabled = instance.xDisabled();


			var area_data = [];
			if (!disabled && state.anno_x_enabled && state.anno_x_areas && stack == state.anno_x_stack) {
				state.anno_x_areas.split("\n").map(function(string) {
					var area = splitLabelAndValue(string);
					if (area.length < 2) return;
					var label = area[0];
					var range = splitRange(area[1]).map(function(label, i) {
						var x_val = xScale(getParsedValue(label, type, chart_layout, "x"));
						if (instance.xExtendByStep()) x_val += xScale.step()/2 * (i == 0 ? -1 : 1);
						return x_val;
					});
					var color = area[2] || state.anno_x_fill_color;
					var facet = area[3] || null;
					if (facet && d.facet_name !== facet) return;
					if (range.length > 1) area_data.push({ label: label, range: range, color: color });
				});
			}

			var areas = select(facet_group).selectAll("g.area.x").data(area_data);
			var areas_enter = areas.enter().append("g").attr("class", "area x")
				.attr("transform", function(d) { return "translate(" + (Math.min(d.range[0], d.range[1]) || margin_left) + ", " + margin_top + ")"; });
			areas_enter.append("rect").attr("height", plot_height)
				.attr("width", function(d) { return (Math.abs(d.range[1] - d.range[0])) || 0; });
			areas_enter.append("text");
			var areas_update = areas.merge(areas_enter);
			areas_update.transition().duration(dur)
				.attr("transform", function(d) { return "translate(" + (Math.min(d.range[0], d.range[1]) || margin_left) + ", " + margin_top + ")"; });
			areas_update.select("rect")
				.attr("width", function(d) { return (Math.abs(d.range[1] - d.range[0])) || 0; })
				.attr("fill", getColor)
				.attr("opacity", state.anno_x_fill_opacity)
				.attr("height", plot_height);
			areas_update.select("text").text(function(d) { return d.label; })
				.attr("fill", state.anno_x_label_color)
				.attr("font-size", getXTextSize())
				.attr("transform", function() {
					var y_shift = v ? 6 : (getXTextSize() * 1.25);
					var ver = a == "bottom" ? (plot_height - y_shift) : (a == "top" ? y_shift : plot_height/2);
					return "translate(" + getXTextSize() * 0.5 + ", " + ver + ") rotate(" + (v ? -90 : 0) + ")";
				})
				.attr("text-anchor", function() {
					if (v) return a == "bottom" ? "start" : (a == "top" ? "end" : "middle");
					return "start";
				})
				.attr("dy", v ? "0.75em" : null);
			areas.exit().remove();
		}

		function updateYLines(facet_group, d, stack) {
			const chart_layout = d.chart_layout;
			let plot_width = Math.max(0, chart_layout.plot_width);
			let a = state.anno_y_label_align;
			let x = chart_layout.margins.left;
			let margin_top = chart_layout.margins.top;
			let yScale = chart_layout.yScale();
			let type = getTypeFromChartLayout(chart_layout.yData());
			let disabled = instance.yDisabled();

			var line_data = [];
			if (!disabled && state.anno_y_enabled && state.anno_y_lines && stack == state.anno_y_stack) {
				state.anno_y_lines.split("\n").forEach(function(string) {
					var line = splitLabelAndValue(string);
					var label = line[0];
					var position = line[1];
					if (position === undefined) return;
					var color = line[2] || state.anno_y_line_color;
					var facet = line[3] || null;
					if (facet && d.facet_name !== facet) return;
					line_data.push({
						label: label,
						y: getParsedValue(position, type, chart_layout, "y"),
						color: color
					});
				});
			}

			var lines = select(facet_group).selectAll("g.line.y").data(line_data);
			var lines_enter = lines.enter().append("g").attr("class", "line y")
				.attr("transform", function(d) { return "translate(" + x + ", " + (yScale(d.y) || 0) + ")"; });
			lines_enter.append("line").attr("width", 0);
			lines_enter.append("text");
			var lines_update = lines.merge(lines_enter);
			lines_update.transition().duration(dur)
				.attr("transform", function(d) { return "translate(" + x + ", " + (yScale(d.y) || 0) + ")"; });
			lines_update.select("line")
				.attr("stroke-dasharray", state.anno_y_line_dash + " " + state.anno_y_line_dash)
				.attr("stroke", getColor)
				.attr("stroke-width", state.anno_y_line_width)
				.attr("x2", plot_width);
			lines_update.select("text").text(function(d) { return d.label; })
				.attr("fill", state.anno_y_line_color)
				.attr("x", a == "left" ? 10 : (a == "right" ? plot_width - 6 : plot_width/2))
				.attr("text-anchor", a == "left" ? "start" : (a == "right" ? "end" : "middle"))
				.attr("dy", function(d) {
					if ((yScale(d.y) || 0) < (getXTextSize() + margin_top)) return "1.25em";
					else return "-0.5em";
				})
				.attr("font-size", getYTextSize());
			lines.exit().remove();
		}

		function updateYAreas(facet_group, d, stack) {
			const chart_layout = d.chart_layout;
			let plot_width = Math.max(0, chart_layout.plot_width);
			let a = state.anno_y_label_align;
			let x = chart_layout.margins.left;
			let yScale = chart_layout.yScale();
			let type = getTypeFromChartLayout(chart_layout.yData());
			let disabled = instance.yDisabled();

			var area_data = [];
			if (!disabled && state.anno_y_enabled && state.anno_y_areas && stack == state.anno_y_stack) {
				state.anno_y_areas.split("\n").map(function(string) {
					var area = splitLabelAndValue(string);
					if (area.length < 2) return;
					var label = area[0];
					var range = splitRange(area[1]).map(function(label, i) {
						var y_val = yScale(getParsedValue(label, type, chart_layout, "y"));
						if (instance.yExtendByStep()) y_val += yScale.step()/2 * (i == 0 ? 1 : -1);
						return y_val;
					});
					var color = area[2] || state.anno_y_fill_color;
					var facet = area[3] || null;
					if (facet && d.facet_name !== facet) return;
					if (range.length > 1) area_data.push({ label: label, range: range, color: color });
				});
			}

			var areas = select(facet_group).selectAll("g.area.y").data(area_data);
			var areas_enter = areas.enter().append("g").attr("class", "area y")
				.attr("transform", function(d) { return "translate(" + x + ", " + Math.min(d.range[0], d.range[1]) + ")"; });
			areas_enter.append("rect").attr("width", plot_width)
				.attr("height", function(d) { return (Math.abs(d.range[1] - d.range[0])); });
			areas_enter.append("text");
			var areas_update = areas.merge(areas_enter);
			areas_update.transition().duration(dur)
				.attr("transform", function(d) { return "translate(" + x + ", " + Math.min(d.range[0], d.range[1]) + ")"; });
			areas_update.select("rect")
				.attr("height", function(d) { return (Math.abs(d.range[1] - d.range[0])); })
				.attr("fill", getColor)
				.attr("opacity", state.anno_y_fill_opacity)
				.attr("width", plot_width);
			areas_update.select("text").text(function(d) { return d.label; })
				.attr("fill", state.anno_y_label_color)
				.attr("x", a == "left" ? 10 : (a == "right" ? plot_width - 6 : plot_width/2))
				.attr("text-anchor", a == "left" ? "start" : (a == "right" ? "end" : "middle"))
				.attr("dominant-baseline", "hanging")
				.attr("dy", "0.33em").attr("font-size", getYTextSize());
			areas.exit().remove();
		}

		function updateHighlights(stack) {
			var axis_highlights = select(".annotations." + stack).selectAll("g.axis-highlights").data(data);
			var axis_highlights_enter = axis_highlights.enter().append("g").attr("class", "axis-highlights")
				.attr("transform", function(d) { return "translate(" + d.facet_offsets.x + ", " + d.facet_offsets.y + ")"; })
				.style("opacity", 0);
			axis_highlights.merge(axis_highlights_enter)
				.transition().duration(instance.animationDuration())
				.style("opacity", 1)
				.attr("transform", function(d) { return "translate(" + d.facet_offsets.x + ", " + d.facet_offsets.y + ")"; })
				.each(function(d) {
					var facet_group = this;
					updateXLines(facet_group, d, stack);
					updateXAreas(facet_group, d, stack);
					updateYLines(facet_group, d, stack);
					updateYAreas(facet_group, d, stack);
				});

			axis_highlights.exit().remove();

			// TODO One problem with the current design is that bar/column charts
			// have discrete scales even when the categories are numerical
			// values. This means that the user can't use fractions to position
			// their highlights more accurately – e.g. if you use 2015.5 on
			// a bar chart you'll get an error rather than a line between 2015
			// and 2016. Might be good to throw a Flourish.console message.
		}


		parseNum = instance.parser();
		updateHighlights("above");
		updateHighlights("below");
	};
}


export { initUpdate };
